/*
 * appcontroller.js
 *
 *  Created on: 08.02.2016
 *      Author: Holger Dahm
 *
 * the module provides all app message, caching as well as internet
 * connection stuff. in addition it provides a "processing" interface
 * that e.g. ensures that no two requests are running simultaneously.
 *
 *  Note: normally no change in this file is necessary - it's generic
 */

// helper lib with some useful operations
var pebblehelpers = require("./pebblehelpers");
// helper lib that converts xml content to json
var xml2js = require("../../../../xml2js/lib/xml2js");



// (default) options used in this module
// Note: call setOptions() before doing anything!
var options;

function setOptions(values) {
	options = values;
}



// local attributes

// flag that indicates that processing is still in progress; the request counter ensures that the flag is
// unset in case 3 requests are blocked in serie
var processing           = false;
var requestConter        = 3;

// app message send properties
var MAX_TRIES            = 3;
var RETRY_TIMEOUT        = 3000;
var APPMESSAGE_TIMEOUT   = 240;

// central queue of all appmessages
var appMessageQueue = [];



/*
 * "processing is ongoing" interace methods
 */

function setProcessing(value) {
	processing = value;

	if (processing) {
		console.log("Start processing on " + pebblehelpers.getPebblePlatform() + " platform.");

		pebblehelpers.detectPebblePlatformChange();
	}
	else {
		// reset counter
        requestConter = 3;
	}
}

function isProcessing() {
    if (requestConter < 1) {
    	setProcessing(false);
    }

    if (!processing) {
    	// nothing remains to do
        return false;
    }

    requestConter--;

    console.log("ERROR: I'm still processing a request - blocked new request. [requestConter: " + requestConter + "]" );
    appMessageQueue.push({
        "message" : {
            "processing" : "processing"
        }
    });
    sendAppMessage(false);

    return true;
}

/*
 * generic method to detect if provider is general/ time dependent requested or not
 * Note: in case provider isn't requested and "send" is true a message is send
 * to pebble automatically
 */

function isRequested(name, show, values, start, stop, send) {
	var noDataString = "";

    if (values == null || values.length == 0) {
    	noDataString = "no " + name + " selected";
    }
    else if (show == 0 || values.length == 0) {
        // nothing requested nothing remains to do
    	noDataString = "no " + name + " mode";
    }
    else if  ((show == 2 && !pebblehelpers.isWeakday()) || !pebblehelpers.nowIsBetween(start, stop)) {
        // out of defined timeframe
    	noDataString = "no " + name + " time";
    }

    if (noDataString.length > 0) {
    	if (send) {
	        appMessageQueue.push({
	            "message" : {
	                "nodata" : noDataString
	            }
	        });
	        sendAppMessage(true);
    	}
        return false;
    }

    return true;
}

/*
 * fill app queue
 * provides in addition temporary caching operations
 */

function pushToLocalStorageAndAppMessages(name, jsonMessage) {
    appMessageQueue.push(jsonMessage);

    // store message temporary
    jsonMessage.timestamp = Math.round(new Date().getTime() / 1000);
    localStorage.setItem(name, JSON.stringify(jsonMessage));
    //console.log(name + ' stored in localStorage (timestamp: ' + jsonMessage.timestamp + '): ' + JSON.stringify(jsonMessage));
}

function pushLocalStorageToAppMessages(name) {
    var message = localStorage.getItem(name);

    if (message != null) {
        //console.log(name + ' in localStorage: ' + message);
        var jsonMessage = JSON.parse(message);

        var now = Math.round(new Date().getTime() / 1000);

        if (now - jsonMessage.timestamp < 2 * 60 * options.reload) { // 2 * reload time
            appMessageQueue.push(jsonMessage);
        }
        else {
            //console.log('ERROR: ' + name + ' out of timeframe!');
        }
    }
    else {
        //console.log('ERROR: ' + name + ' not stored in localStorage!');
    }
}

/*
 * send all messages currently in app queue to pebble
 */

function resendAppMessage(appMessage, finalizeRequest) {
    if (appMessage.numTries > MAX_TRIES) {
        console.log("Failed sending AppMessage. Bailing. " /*+ JSON.stringify(appMessage.message)*/);

        // nothing remains to do
        return;
    }

    Pebble.sendAppMessage(appMessage.message,
        function(e) {
            //console.log("AppMessage send to pebble! " + JSON.stringify(appMessage.message));

            setTimeout(function() {
            	sendAppMessage(finalizeRequest);
            }, APPMESSAGE_TIMEOUT);
        },
        function(e) {
            console.log("Failed sending AppMessage. Error: " + e.data.error.message);
            appMessage.numTries++;

            setTimeout(function() {
            	resendAppMessage(appMessage, finalizeRequest);
            }, RETRY_TIMEOUT);
        }
    );
}

function sendAppMessage(finalizeRequest) {
    if (appMessageQueue.length == 0) {
        // nothing remains to do
        if (finalizeRequest) {
        	setProcessing(false);
        }
        return;
    }

    var appMessage = appMessageQueue.shift();

    // set default values
    appMessage.numTries = appMessage.numTries || 0;

    resendAppMessage(appMessage, finalizeRequest);
}

/*
 * methods to get XML or JSON resources from the internet
 */

function startRequest(counter, url, processingFunction, errorFunction) {
    if (counter < 1) {
        // that's it - no success :(
        console.log("Last try - Can't get HTTP page.");// " + url);
        if (typeof errorFunction === "undefined") {
            appMessageQueue.push({
                "message" : {
                    "error" : "no internet"
                }
            });
            sendAppMessage(true);
        }
        else {
            errorFunction();
        }
        return;
    }
    //console.log("Get HTTP page: " + url);
    //console.log(counter + " tries remaining");

    var request = new XMLHttpRequest();
    request.open("GET", url, true);
    request.onload = function(e) {
        if (request.readyState == 4 && request.status == 200 && request.responseText) {
            try {
                processingFunction(request.responseText);
                sendAppMessage(true);
            }
            catch (error) {
                console.log("Error during response processing!\n" + error.stack);
                setTimeout(function() {
                    startRequest(--counter, url, processingFunction, errorFunction);
                }, RETRY_TIMEOUT);
            }
        }
        else {
            console.log("Got HTML reponse but with error code! (readyState: " + request.readyState + "; status: " + request.status + ")");
            setTimeout(function() {
                startRequest(--counter, url, processingFunction, errorFunction);
            }, RETRY_TIMEOUT);
        }
    };
    request.onerror = function(e) {
        console.log("Got error during http request! (status: " + e.target.status + ")");
        setTimeout(function() {
            startRequest(--counter, url, processingFunction, errorFunction);
        }, RETRY_TIMEOUT);
    };
    request.send(null);
}

function startJSONRequest(url, parseFunction, errorFunction) {
    startRequest(MAX_TRIES, url,
        function(responseText) {
            parseFunction(JSON.parse(responseText));
        },
        errorFunction
    );
}

function startXMLRequest(url, parseFunction, errorFunction) {
    startRequest(MAX_TRIES, url,
        function(responseText) {
            xml2js.parseString(responseText, parseFunction);
        },
        errorFunction
    );
}



/*
 * module interace
 */

module.exports = {
		// (default) options used in this module
		setOptions : setOptions,

		// module providing some helpful function
		pebblehelpers : pebblehelpers,

		// "processing is ongoing" interace methods
		setProcessing : setProcessing,
		isProcessing : isProcessing,

		// generic method to detect if provider is general/ time dependent requested or not
		isRequested : isRequested,

		// fill (cached) app queue
		appMessageQueue : appMessageQueue,

		pushToLocalStorageAndAppMessages : pushToLocalStorageAndAppMessages,
		pushLocalStorageToAppMessages : pushLocalStorageToAppMessages,

		// send all messages currently in app queue to pebble
		sendAppMessage : sendAppMessage,

		// methods to get XML or JSON resources from the internet
		startJSONRequest : startJSONRequest,
		startXMLRequest : startXMLRequest,
};
